; ddr.asm - Interactive debugger / monitor
;
; Copyright (C) 1999 Thomas Novelli
;	Retro is distributed under the terms of the GNU LGPL, version 2 or
;	greater. (For the text of the license, see the file COPYING.LIB)
;
; Progress:
;	3/18/99 Got a simple debugger running. Lets you run any command in
;		the symbol table (i.e., anything declared as "global" in a
;		module). Lets you pass string arguments to the command,
;		including quoted strings.
;
;	3/19/99 Got the bugs worked out. Inputs decimal, hex, octal & binary
;		numbers, including negative number.
;
; To do:
;	* Use the ODB.. the old symbol table is gone :)
;
;	* colon-delimited object identifiers, like this:
;		> tty:print "Hello world"
;
;	Let you modify variables, preferably with something like this:
;		> attr=$0C
;
;	Allow chaining of commands:
;		> attr=$0C ; print "This is bright red" ; cr
;

%include "kernel.mac"
%include "macros.mac"
%include "kbd.inc"
%include "../memory.inc"

extern	attr, _print

	SECTION .text

;init:	ret


	global ddr
ddr:
;
; Why DDR? It's short. It sounds good. DDT for Retro :)
; (What the hell does DDT stand for, anyway?)
;
	mov	[attr], byte 0x07
	print	start_msg
.prompt
	mov	[attr], byte 0x0C
	print	prompt
	mov	[attr], byte 0x0F
	call	readline		; -- eax
	mov	[attr], byte 0x07
	call	lookup			; eax -- eax esi

	or	eax, eax	;Not found? Back to prompt..
	jz	.prompt

	; Parse arguments & push them on the stack
	xor	ecx, ecx	;ecx = # of items pushed on stack
.parse
	call	parse_arg	; esi -- esi edi
	or	esi, esi	;Done parsing line?
	jz	.call
	push	edi		;Push argument on stack
	inc	ecx
	jmp short .parse
.call
	call	eax		;Execute command
	shl	ecx, 2
	add	esp, ecx	;Remove arguments from stack
	jmp .prompt


parse_arg:
;
; Parameters:
;	ESI = address of string
;
; Returns:
;	ESI = remainder of string (0 if at end)
;	EDI = result (either a number, or a string's address)
;
; Number formats: !233=decimal, $233=hex, %101=binary, &233=octal
;	default is decimal (i.e., 233 is treated as decimal)
;
	mpush	eax, ebx, ecx, edx
	mov	edi, esi	;Return a string, by default

	dec	esi		;Skip leading spaces - quick hack!
.skip_space
	inc	esi
	cmp	[esi], byte 32
	je	.skip_space

	cmp	[esi], byte '"'		;Quoted string?
	je near .quote

	;Figure out base (if it's a number)
	mov	[base], dword 10	; Default = decimal
	cmp	[esi], byte '$'		; $ = hex
	jne	.not_hex
	mov	[base], dword 16
	inc	esi			;Skip past $
	jmp	.number
.not_hex
	cmp	[esi], byte '%'		; % = binary
	jne	.not_binary
	mov	[base], dword 2
	inc	esi			;Skip past %
	jmp	.number
.not_binary
	cmp	[esi], byte '!'		; ! = decimal (the default)
	jne	.not_decimal
	inc	esi			;Skip past !
	jmp	.number
.not_decimal
	cmp	[esi], byte '&'		; & = octal
	jne	.number
	mov	[base], dword 8
	inc	esi

.number		;TRY to convert to a number
	;
	; Algorithm:
	;	Check for - and set flag
	;	Read a digit
	;	Multiply accumulator by base & add digit
	;
	xor	eax, eax		;eax = Counter (the number)
	xor	ebx, ebx		;clear ebx (ebx is added to eax)
	xor	ecx, ecx		;ecx = # of digits
	mov	[sign], al		;sign = 0 (-1 means negative)
	
	cmp	[esi], byte '-'		;Negative?
	jne	.loop
	dec	byte [sign]
	inc	esi

.loop
	mov	bl, [esi]	;Fetch next digit & translate it
	cmp	bl, '0'
	jb	.number_end		;still here? bl >= '0'
	cmp	bl, '9'			;'0' <= bl <= '9'
	jbe	.decimal
	cmp	bl, 'A'
	jb	.number_end		;still here? bl >= 'A'
	cmp	bl, 'Z'			;'A' <= bl <= 'Z'
	jbe	.upr
	cmp	bl, 'a'
	jb	.number_end
	cmp	bl, 'z'
	jbe	.lwr
	jmp short .number_end

.lwr
	sub	bl, 'a'-10		;Convert lowercase letter to #
	jmp short .check_base
.upr
	sub	bl, 'A'-10		;Convert uppercase letter to #
	jmp short .check_base
.decimal
	sub	bl, '0'			;Convert decimal digit to #
.check_base
	cmp	bl, byte [base]		;Make sure it's not above the base
	jae	.number_end

	mul	dword [base]		;Add to the counter
	add	eax, ebx
	inc	ecx			;increment digit-counter
	inc	esi			;Get next digit
	jmp short .loop

.number_end
	or	ecx, ecx		;Any digits found?
	jz	.not_number
	cmp	byte [sign], 0		;Negative? flip sign.
	jz	.not_neg
	neg	eax
.not_neg
	mov	edi, eax		;return the number
	jmp short .end

.not_number
	mov	esi, edi	;Reset to beginning of string

.strip		;Parse out a space-delimited string
	mov	al, [esi]
	cmp	al, 32		;Whitespace?
	jz	.space
	cmp	al, 8
	jz	.space
	or	al, al		;End of input string?
	jnz	.next		; no - get next byte

	cmp	esi, edi	;First char is 0?
	jnz	.end
	xor	esi, esi	;Then, return ESI=0 (end of input)
.end
	mpop	eax, ebx, ecx, edx
	ret

.space
	cmp	esi, edi	;Are we at the beginning of the line?
	jnz	.end_word	; no: End of word
	inc	edi		; yes: skip the initial space
.next
	inc	esi
	jmp short .strip
.end_word			; no: it's a space after a word; we're done
	mov	[esi], byte 0
	inc	esi
	jmp short .end

.quote	;Skip the first " and read till next " or 0
	inc	edi	;Skip first quote
.quote_scan
	inc	esi
	cmp	[esi], byte '"'
	jz	.end_quote
	cmp	[esi], byte 0
	jnz	.quote_scan
.end_quote
	mov	[esi], byte 0
	jmp short .end



lookup:
;
; Look up the first word in the string pointed to by EAX.
;
; Parameters:
;	EAX = input string
;
; Returns:
;	EAX = address (0 if not found)
;	ESI = remainder of input string
;
	mpush	ecx, edx

	cmp	[eax], byte 0	;Empty string? "Never mind"
	jz	.abort

	mov	[buffer], eax		;Save address of buffer
	mov	eax, SYMTAB_BASE	;eax = current entry
.search
	mov	esi, [buffer]
	mov	eax, [eax]	;Get next entry
	or	eax, eax	;At end of symtab yet?
	jz	.no_luck
	mov	ecx, eax	;ecx = name string (in symtab)
	add	ecx, 8
.scan
	inc	esi
	inc	ecx
	mov	dl, [esi-1]
	call	.convert	;Convert whitespace to 0
	cmp	dl, [ecx-1]
	jnz	.search		;Different? "NEXT!!"
	or	dl, dl		;End of word, and still no differences?
	jnz	.scan
.match
	mov	eax, [eax+4]	;Return address of routine
.end
	mpop	ecx, edx
	ret
.no_luck
	print	not_found
.abort
	xor	eax, eax	;Return 0
	jmp short .end

.convert
; If DL is a whitespace character, convert it to 0
	cmp	dl, 32		;SPACE = end of word.
	jz	.convert_ok
	cmp	dl, 9		;TAB is also whitespace
	jz	.convert_ok
.convert_end
	ret
.convert_ok
	xor	dl,dl
	ret



	SECTION .data
start_msg	db	10,"DDR (Retro debugger)",10
		db	"Type 'help' for commands",10,0
prompt		db	"> ",0
not_found	db	"No comprendo",10,0


	SECTION .bss
buffer		resd 1
base		resd 1
sign		resd 1
